@import url("{{ pathto('_static/fontawesome/css/fa-brands.min.css', 1) }}");

.console-block {
  text-align: right;
}

.console-block *:before,
.console-block *:after {
  box-sizing: border-box;
}

.console-block > section {
  display: none;
  text-align: left;
}

.console-block > input.c-tab-unix,
.console-block > input.c-tab-win {
  display: none;
}

.console-block > label {
  display: inline-block;
  padding: 4px 8px;
  font-weight: normal;
  text-align: center;
  color: #bbb;
  border: 1px solid transparent;
  font-family: fontawesome;
}

.console-block > input:checked + label {
  color: #555;
  border: 1px solid #ddd;
  border-top: 2px solid #ab5603;
  border-bottom: 1px solid #fff;
}

.console-block > .c-tab-unix:checked ~ .c-content-unix,
.console-block > .c-tab-win:checked  ~ .c-content-win {
  display: block;
}

.console-block pre {
  margin-top: 0px;
}
